#pragma once

#include <iostream>
#include <string_view>
#include <source_location>
#include <cxxabi.h>
#include <libunwind.h>
#include <dlfcn.h>
#include <cstring>
#include <memory>
#include <sstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>

// ANSI color codes
namespace colors {
    constexpr const char* reset = "\033[0m";
    constexpr const char* red = "\033[31m";
    constexpr const char* green = "\033[32m";
    constexpr const char* yellow = "\033[33m";
    constexpr const char* blue = "\033[34m";
    constexpr const char* magenta = "\033[35m";
    constexpr const char* cyan = "\033[36m";
}

// Simple RAII wrapper for file descriptor
class FileDescriptor {
    int fd_ = -1;
public:
    explicit FileDescriptor(int fd) : fd_(fd) {}
    ~FileDescriptor() { if (fd_ != -1) close(fd_); }
    operator int() const { return fd_; }
    FileDescriptor(const FileDescriptor&) = delete;
    FileDescriptor& operator=(const FileDescriptor&) = delete;
};

struct SourceInfo {
    std::string function;
    std::string file;
    int line = 0;
};

SourceInfo get_source_info(const char* executable, void* addr) {
    char cmd[512];
    snprintf(cmd, sizeof(cmd), "addr2line -f -e %s %p 2>/dev/null", executable, addr);
    
    FILE* pipe = popen(cmd, "r");
    if (!pipe) return {};
    
    SourceInfo info;
    char func_buf[1024] = {0};
    char file_buf[1024] = {0};
    int line = 0;
    
    // Read function name
    if (fgets(func_buf, sizeof(func_buf), pipe)) {
        // Remove trailing newline
        size_t len = strlen(func_buf);
        if (len > 0 && func_buf[len-1] == '\n') {
            func_buf[len-1] = '\0';
        }
        info.function = func_buf;
        
        // Demangle the function name
        int status = 0;
        char* demangled = abi::__cxa_demangle(func_buf, nullptr, nullptr, &status);
        if (demangled) {
            info.function = demangled;
            free(demangled);
        }
    }
    
    // Read file and line number
    if (fgets(file_buf, sizeof(file_buf), pipe)) {
        char* colon = strchr(file_buf, ':');
        if (colon) {
            *colon = '\0';
            info.line = atoi(colon + 1);
            
            // Extract just the filename
            const char* slash = strrchr(file_buf, '/');
            info.file = (slash ? slash + 1 : file_buf);
        }
    }
    
    pclose(pipe);
    return info;
}

void print_stacktrace() {
    unw_cursor_t cursor;
    unw_context_t context;
    unw_word_t pc;
    char exe_path[1024] = {0};
    bool found_main = false;
    std::vector<SourceInfo> frames;

    // Get the path to the current executable
    ssize_t count = readlink("/proc/self/exe", exe_path, sizeof(exe_path) - 1);
    if (count == -1) {
        strcpy(exe_path, "[unknown]");
    } else {
        exe_path[count] = '\0';
    }

    // Initialize cursor to current frame
    unw_getcontext(&context);
    unw_init_local(&cursor, &context);

    // First pass: collect frames until we find main
    while (unw_step(&cursor) > 0) {
        if (unw_get_reg(&cursor, UNW_REG_IP, &pc) != 0 || pc == 0) {
            break;
        }

        SourceInfo info = get_source_info(exe_path, reinterpret_cast<void*>(pc));
        
        // Stop collecting after main()
        if (!info.function.empty() && 
            (info.function.find("main") != std::string::npos || 
             info.function.find("__libc_start_main") != std::string::npos)) {
            found_main = true;
            frames.push_back(info);
            break;
        }
        
        frames.push_back(info);
    }

    // Print the collected frames in reverse order (most recent last)
    std::cerr << colors::yellow << "Stack trace (most recent call last):" << colors::reset << "\n";
    
    for (size_t i = frames.size(); i-- > 0; ) {
        const auto& frame = frames[i];
        std::cerr << "  ";  // Indent each line
        
        if (frame.function.empty()) {
            std::cerr << colors::red << "[unknown function]" << colors::reset;
        } else {
            std::cerr << colors::green << frame.function << colors::reset;
        }
        
        if (!frame.file.empty() && frame.line > 0) {
            std::cerr << " at " << colors::blue << frame.file << colors::reset 
                     << ":" << colors::magenta << frame.line << colors::reset;
        }
        
        std::cerr << "\n";
    }
    
    if (!found_main) {
        std::cerr << "  " << colors::yellow << "[truncated - main() not found]" << colors::reset << "\n";
    }
}

void assert_failed(
    bool condition,
    std::string_view message,
    std::source_location location = std::source_location::current()
) {
    if (!condition) {
        std::cerr << "Assertion failed at " << location.file_name() << ":" << location.line() << ": " 
                  << location.function_name() << "(): " << message << "\n";
        print_stacktrace();
        std::abort();
    }
}

#define ASSERT(condition, message) assert_failed(condition, message, std::source_location::current())
